@@ -16,36 +16,43 @@
 
 -module(rabbit_tests_event_receiver).
 
--export([start/1, stop/0]).
+-export([start/3, stop/0]).
 
 -export([init/1, handle_call/2, handle_event/2, handle_info/2,
          terminate/2, code_change/3]).
 
-start(Pid) ->
-    gen_event:add_handler(rabbit_event, ?MODULE, [Pid]).
+-include("rabbit.hrl").
+
+start(Pid, Nodes, Types) ->
+    Oks = [ok || _ <- Nodes],
+    {Oks, _} = rpc:multicall(Nodes, gen_event, add_handler,
+                             [rabbit_event, ?MODULE, [Pid, Types]]).
 
 stop() ->
     gen_event:delete_handler(rabbit_event, ?MODULE, []).
 
 %%----------------------------------------------------------------------------
 
-init([Pid]) ->
-    {ok, Pid}.
+init([Pid, Types]) ->
+    {ok, {Pid, Types}}.
 
-handle_call(_Request, Pid) ->
-    {ok, not_understood, Pid}.
+handle_call(_Request, State) ->
+    {ok, not_understood, State}.
 
-handle_event(Event, Pid) ->
-    Pid ! Event,
-    {ok, Pid}.
+handle_event(Event = #event{type = Type}, State = {Pid, Types}) ->
+    case lists:member(Type, Types) of
+        true  -> Pid ! Event;
+        false -> ok
+    end,
+    {ok, State}.
 
-handle_info(_Info, Pid) ->
-    {ok, Pid}.
+handle_info(_Info, State) ->
+    {ok, State}.
 
-terminate(_Arg, _Pid) ->
+terminate(_Arg, _State) ->
     ok.
 
-code_change(_OldVsn, Pid, _Extra) ->
-    {ok, Pid}.
+code_change(_OldVsn, State, _Extra) ->
+    {ok, State}.
 
 %%----------------------------------------------------------------------------
